#braille_cell.py    27Nov2022
from select_trace import SlTrace

"""
Basis for Braille cell (6 dot pattern)
"""
class BrailleCell:
    """ braille cell info augmented for analysis
    """
    # Marking type
    MARK_UNMARKED = 1
    MARK_SELECTED = 2
    MARK_TRAVERSED = 3

    # Colors we use
    color_for_character = {
        "r": "red",
        "o": "orange",
        "y": "yellow",
        "g": "green",
        "b": "blue",
        "i": "indigo",
        "v": "violet",
        "w": "white",
        "p": "pink",
        "l": "lavender",
         
        }
    
    dots_for_character = {
        " ": (),    # blank
        "a": (1),
        "b": (1,2),
        "c": (1,4),
        "d": (1,4,5),
        "e": (1,5),
        "f": (1,2,5),
        "g": (1,2,4,5),
        "h": (1,2,5),
        "i": (2,4),
        "j": (2,4,5),
        "k": (1,3),
        "l": (1,2,3),
        "m": (1,3,4),
        "n": (1,3,4,5),
        "o": (1,3,5),
        "p": (1,2,3,4),
        "q": (1,2,3,4,5),
        "r": (1,2,3,5),
        "s": (2,3,4),
        "t": (2,3,4,5),
        "u": (1,3,6),
        "v": (1,2,3,6),
        "w": (2,4,5,6),
        "x": (1,3,4,6),
        "y": (1,3,4,5,6),
        "z": (1,3,5,6),
        }

    @classmethod
    def braille_for_letter(cls, c):
        """ convert letter to dot number seq
        :c: character
        :returns: dots tupple (1,2,3,4,5,6)
        """
        if c not in cls.dots_for_character:            c = " " # blank
        dots = cls.dots_for_character[c]
        return dots

    @classmethod
    def ch2color(cls, ch):
        """ Convert character to recognized color name
        :ch: character, first character of color name
        :returns: color name, black if none
        """
        if ch in cls.color_for_character:
            return cls.color_for_character[ch]
        
        return "black"  # acceptible tkinter but not in our group
        
    @classmethod
    def color_str(cls, color):
        """ convert turtle colors arg(s) to color string
        :color: turtle color arg
        """
        color_str = color
        if isinstance(color_str,tuple):
            if len(color_str) == 1:
                color_str = color_str[0]
            else:
                color_str = "pink"  # TBD - color tuple work
        return color_str
    
    @classmethod
    def tuple_to_braille_cell(cls, tp):
        """ Create BrailleCell from (ix,iy,color)
        if given BrailleCell return it unchanged
        """
        if isinstance(tp, BrailleCell):
            return tp
        
        if len(tp) < 3:
            SlTrace.lg(f"tuple_to_braille_cell({tp})")
            tp = (0,0,"red")
        ix,iy,color = tp
        bc = BrailleCell(ix=ix, iy=iy, color=color)
        return bc
    
    
    def __init__(self, dots=None,
                 color=None, color_bg=None,
                 ix=0, iy=0,
                 mtype=None,
                 visible=True,
                 points=None):
        """ setup braille cell
        :dots: list of set dots default: none - blank
        :color: color str or tuple
        :ix: cell index(from 0) from left side
        :iy: cell index from bottom
        :mtype: marked type
                default: MARK_UNSELECTED
        :visible: cell is visible default: True
        :points: initial set of points, if any
            default: empty
        """
        self.ix = ix    # Include to make self sufficient
        self.iy = iy
        if mtype is None:
            mtype = BrailleCell.MARK_UNMARKED
        self.mtype = mtype
        self._visible = visible
        if color is None or len(color)==0:
            color = "black"
        if dots is None:
            char = color[0]
            if char in self.dots_for_character:
                dots = self.dots_for_character[char]
            else:
                dots = self.dots_for_character[" "]
        self.dots = dots
        if color_bg is None: 
            color_bg = "white"
        self._color = color
        self._color_bg = color_bg
        if points is None:
            points = set()
        self.points = points
        self.canv_items = []        # canvas items
        self.pi_number = None       # perimeter traverse number, if one 
        self.pi_type = ""           # o-outside, s-surrounding, i-inside
        self.comp_item = None       # graphics item, if one

    def __str__(self):
        st = f"BCell: [{self.ix},{self.iy}]"
        if self._color is not None:
            st += " " + self._color
        if self._color_bg is not None:
            st += " bg:" + self._color_bg
        return st
        
    
    def braille_for_color(self, color):
        """ Return dot list for color
        :color: color string or tuple
        :returns: list of dots 1,2,..6 for first
                letter of color
        """
        
        if color is None:
            color = self._color
        if color is None:
            color = ("black")
        color = self.color_str(color)
        c = color[0]
        dots = self.braille_for_letter(c)
        return dots

    def color_string(self, color=None):
        """ convert turtle colors arg(s) to color string
        :color: turtle color arg
        """
        if color is None:
            color = self._color
        return self.color_str(color)
    
    def color_cell(self, color=None):
        """ Color cell
        :color: turtle color
        """
        if color is None:
            color = self._color
        dots = self.braille_for_color(color=color)
        self.dots = dots
        self._color = color

    def is_visible(self):
        """Check if visible
        :returns: True iff cell is visible
        """
        return self._visible
    
    def set_ix(self, ix):
        """Set/modify ix index
        :ix: x index 
        """
        self.ix = ix
    
    def set_iy(self, iy):
        """Set/modify iy index
        :iy: y index 
        """
        self.iy = iy
        


# trace_control_window.py 27Jan2020
import sys
from tkinter import *
import atexit

import time

from select_trace import SlTrace
from crs_funs import str2val

class TraceControlWindow(Toplevel):
    def __init__(self, tcbase=None, change_call=None):
        """ Trace flag dictionary
        :tcbase: - parent - call basis must have tc_destroy to be called if we close
        :change_call: - call with change flag, value, if present
        """
        self.flag_by_cb = {}             # Dictionary hashed on cb widget
        self.data_by_flag = {}
        self.standalone = False      # Set True if standalone operation
        if tcbase is None:
            SlTrace.lg("Standalone TraceControlWindow")
            root = Tk()
            frame = Frame(root)
            frame.pack()
            self.standalone = True
            root.withdraw()
        self.tcbase = tcbase
        
        self.change_call = change_call
                    
        self.tc_mw = Toplevel()
        tc_x0 = 800
        tc_y0 = 100
        tc_w = 200
        tc_h = 200
        tc_geo = "%dx%d+%d+%d" % (tc_w, tc_h, tc_x0, tc_y0)
        self.tc_mw.geometry(tc_geo)
        self.tc_mw.title("Trace")
        top_frame = Frame(self.tc_mw)
        self.tc_mw.protocol("WM_DELETE_WINDOW", self.delete_tc_window)
        top_frame.pack(side="top", fill="both", expand=False)
        self.top_frame = top_frame
        tc_all_button = Button(master=self.top_frame, text="SET ALL", command=self.select_all)
        tc_all_button.pack(side="left", fill="both", expand=False)
        tc_none_button = Button(master=self.top_frame, text="NONE", command=self.select_none)
        tc_none_button.pack(side="left", fill="both", expand=False)
        tc_bpt_button = Button(master=self.top_frame, text="BPT", command=self.breakpoint)
        tc_bpt_button.pack(side="left", fill="both", expand=False)
        self.show_list_which = "ALL"
        self.show_list_variable = StringVar()
        self.show_list_variable.set("Show SET")
        tc_show_button = Button(master=self.top_frame, textvariable=self.show_list_variable,
                                command=self.show_list)
        tc_show_button.pack(side="left", fill="none", expand=False)
        self.tc_show_button = tc_show_button
        tc_frame = Frame(self.tc_mw)
        tc_frame.pack(side="top", fill="both", expand=True)
        self.tc_frame = tc_frame
        self.tc_text_frame = None           # So we can destroy / resetup in create_flags_region
        self.flags_text = None
        self.sb = None
        self.create_flags_region(flags=SlTrace.getAllTraceFlags())
        if SlTrace.trace("trace_flags"):
            self.list_ckbuttons()
        
    def create_flags_region(self, flags=None):
        if self.tc_text_frame is not None:
            self.tc_text_frame.pack_forget()
            self.tc_text_frame.destroy()
            self.tc_text_frame = None
            
        if self.sb is not None:
            self.sb.destroy()
        self.update()                           # Show progress
        self.tc_text_frame = Frame(self.tc_frame)
        self.tc_text_frame.pack(side="top", fill="both", expand=True)

        self.start = 0
        self.sb = Scrollbar(master=self.tc_text_frame, orient="vertical")
        max_width = 5
        min_height = 10
        t_height = min_height
        max_height = 20
        nfound = 0
        for flag in flags:
            val = SlTrace.getLevel(flag)
            width = len(flag)
            if callable(val):
                val_len = 5
            elif type(val) == bool:
                val_len = 2
            else:
                val_len = len(str(val))
            width += val_len
            if width > max_width:
                max_width = width
            nfound += 1
        win_width = max_width
        if nfound < min_height:
            t_height = min_height
        if nfound > max_height:
            t_height = max_height
            
        text = Text(self.tc_text_frame, width=win_width, height=t_height,
                    yscrollcommand=self.sb.set,
                    state=DISABLED)
        self.sb.config(command=text.yview)
        self.sb.pack(side="right",fill="y")
        text.pack(side="top", fill="both", expand=True)
        self.update()                           # Show progress
        self.flag_by_cb = {}             # Dictionary hashed on cb widget
        self.data_by_flag = {}
        for flag in sorted(flags):
            level = SlTrace.getLevel(flag)
            if type(level) == bool:
                var = BooleanVar()
                var.set(level)
                ####fmt_text = "%-*s" % (max_width, flag)
                fmt_text = flag
                cb = Checkbutton(text, text=fmt_text, padx=0, pady=0, bd=0, variable = var, bg="white")
                self.flag_by_cb[cb] = flag
                self.data_by_flag[flag] = (cb, flag, var)
                text.config(state=NORMAL)
                text.window_create("end", window=cb)
                text.insert("end", "\n")
                text.config(state=DISABLED)
                cb.bind("<Button-1>", self.select_button)
            elif type(level) == int or type(level) == float or type(level) == str:
                if type(level) == int:
                    var = IntVar()
                elif type(level) == float:
                    var = DoubleVar()
                elif type(level) == str:
                    var = StringVar()
                var.set(level)
                var_width = len(str(level)) + 2
                text_var = StringVar()
                text_var.set(f"{level:{var_width}}")
                ####fmt_text = "%-*s" % (max_width, flag)
                fmt_text = flag
                ent = Entry(text, width=var_width, textvariable=text_var)
                self.flag_by_cb[ent] = flag
                self.data_by_flag[flag] = (ent, flag, text_var)     # field is what we need
                text.config(state=NORMAL)
                text.window_create("end", window=ent)
                text.insert("end", fmt_text)
                text.insert("end", "\n")
                text.config(state=DISABLED)
                ent.bind("<Return>", self.enter_entry)
            elif callable(level):
                fmt_text = "%-*s" % (max_width, flag)
                btn = Button(text, text=flag, command=level)
                self.flag_by_cb[btn] = flag
                self.data_by_flag[flag] = (btn, flag, None)     # field is what we need
                text.config(state=NORMAL)
                text.window_create("end", window=btn)
                text.insert("end", "\n")
                
            ###cb.pack()
        self.update()                           # Show progress
        if self.standalone:
            atexit.register(self.on_exit)
            self.update_loop()

    def on_exit(self):
        """ Close down window on program exit
        """
        SlTrace.lg("Closing down Trace Control Window")
        self.delete_tc_window()
        
    def update_loop(self):
        """ continue repeated tk.update() calls
        to enable window operation
        """
        loop_time = 50          # Loop recall time (msec)
        self.update()
        self.tc_mw.after(loop_time, self.update_loop)

    def update(self):
        if self.tc_mw is not None and self.tc_mw.winfo_exists():
            self.tc_mw.update()
        
    def sleep(self, sec):
        """ "sleep" for a number of sec
        without stoping tkinter stuff
        :sec: number of milliseconds to delay before returning
        """
        if self.tc_mw is None:
            return
        
        self.update()             # Insure at least one update
        now = time.time()
        end_time = now + sec
        while time.time() < end_time:
            if self.tc_mw is None:
                return
            
            self.update()
        return
    
    def mainloop(self):
        self.tc_mw.mainloop()
        
    def delete_tc_window(self):
        """ Process Trace Control window close
        """
        if self.tc_mw is not None:
            self.tc_mw.destroy()
            self.tc_mw = None
        
        if self.tcbase is not None and hasattr(self.tcbase, 'tc_destroy'):
            self.tcbase.tc_destroy()

    def show_list(self):
        """ Select buttons to show
        """
        if self.show_list_which == "ALL":
            self.show_list_which = "JUST_SET"
            self.show_list_variable.set("Show ALL")
            just_flags = []
            for flag in SlTrace.getTraceFlags():
                val = SlTrace.getLevel(flag)
                if type(val) != bool:
                    just_flags.append(flag) # Include all non-boolean
                elif val:
                    just_flags.append(flag)
        else:
            self.show_list_which = "ALL"
            self.show_list_variable.set("Show SET")
            just_flags = SlTrace.getTraceFlags()
            
        self.create_flags_region(just_flags)
        
    def select_all(self):
        """ Select all known trace flags
        """
        for flag in sorted(SlTrace.getTraceFlags()):   # In display order
            if type(self.getLevel(flag)) == bool:
                self.set_trace_level(flag, True)


    def select_none(self):
        """ Select all known trace flags
        """
        for flag in sorted(SlTrace.getTraceFlags()):   # In display order
            if type(self.getLevel(flag)) == bool:
                self.set_trace_level(flag, False)

    def getLevel(self, flag):
        return SlTrace.getLevel(flag)
    
    def breakpoint(self):
        """ Force immediate breakpoint - enter debugger
        """
        import pdb
        SlTrace.lg("Breakpoint")
        pdb.set_trace()
                
    def enter_entry(self, event):
        flag = self.flag_by_cb[event.widget]
        old_level = SlTrace.getLevel(flag)
        _, flag, text_var = self.data_by_flag[flag]
        entry_text = text_var.get()
        new_level = str2val(entry_text, old_level)        
        self.set_trace_level(flag, new_level, change_cb=False)  # No select for Entry
                
    def select_button(self, event):
        flag = self.flag_by_cb[event.widget]
        cb, flag, var = self.data_by_flag[flag]
        val = SlTrace.getLevel(flag)        # Variable doesn't seem to work for us
        val = not val                           # Keep value in strace
        self.set_trace_level(flag, val, change_cb=False)  # CB already set
        
        
    def set_trace_level(self, flag, val, change_cb=True):
        """ Set trace level, changing Control button if requested
        :flag: - trace flag name
        :val: - value to set
        :change_cb: True(default) appropriately change the control
        """
        if flag not in self.data_by_flag:
            SlTrace.lg("set_trace_level(%s,%d) - flag has no check button" % (flag, val))
            return
         
        cb, flag, var = self.data_by_flag[flag]        
        if cb is None:
            SlTrace.lg("set_trace_level(%s,%d) - flag None check button" % (flag, val))
            return

        if change_cb and hasattr(cb, "select"):
            if val != 0:
                cb.select()
            else:
                cb.deselect()
                    
        SlTrace.lg("flag=%s, var=%s, val=%s" %(flag, var, val), "trace_flags")
        SlTrace.setLevel(flag, val)
            
        if self.change_call is not None:
            self.change_call(flag, val)


    def list_ckbuttons(self):
        cb_flags = sorted(self.data_by_flag.keys())
        for flag in cb_flags:
            var = self.data_by_flag[flag][2]
            SlTrace.lg(f"flag={flag} var={var} val={var.get()}")

if __name__ == '__main__':
    def report_change(flag, val, cklist=None):
        SlTrace.lg("changed: %s = %d" % (flag, val))
        new_val = SlTrace.getLevel(flag)
        SlTrace.lg("New val: %s = %d" % (flag, new_val))
        if cklist is not None:
            cklist.list_ckbuttons()
    
    root = Tk()
    SlTrace.set_mw(root)
    ###frame = Frame(root)
    ###frame.pack()
    SlTrace.setProps()
    SlTrace.setFlags("flag1=t,flag2=f,flag3=t,flag4=10, flag5=20. flag6=abc")
    threshold = 5
    SlTrace.setLevel("tint1", 5)
    end_level = 100
    quit_set = False
    def our_quit(flag=None):
        """ Test for traceButton
        :flag: flag arg
        """
        global quit_set
        if flag is None:
            flag = "TBD"
        SlTrace.lg(f"our_quit({flag})")
        SlTrace.report("Quitting Program")
        root.destroy
        quit_set = True
        sys.exit()

        
    def test_int(flag, level=threshold, default=None):
        SlTrace.lg(f"Set {flag} to over {end_level} to quit")
        if SlTrace.trace(flag, threshold):
            SlTrace.lg(f"{flag} = {SlTrace.trace(flag)} >= {level}")
        else:
            SlTrace.lg(f"{flag} = {SlTrace.trace(flag)} < {level}")
        if quit_set or SlTrace.trace("tint1") > end_level:
            SlTrace.lg("Manual quit")
            sys.exit()
        SlTrace.traceButton("quit", our_quit)
                    
    app = TraceControlWindow(change_call=report_change)
    
    for i in range(end_level):
        test_int("tint1", default=2)
        app.sleep(1)
        
        
    SlTrace.lg("End of test")


#wx_audio_draw_window.py 24May2023  crs, from audio_draw_window.py
#                        12Dec2022  crs, from audio_window.py
#                        08Nov2022  crs, Author
"""
Remodeled from tkinter base to use wxPython to facilitate
better support with screen readers such as NVDA and JAWS

Provide simple drawing graphical window with audio feedback
to facilitate use and examination by the visually impaired.
TBD: Using wxPython to facilitate cursor movement within screen
Adapted from audio_window to concentrate on figure drawing
as well as presentation.
"""
import sys
import os
import datetime
import re

import wx
import traceback

from select_trace import SlTrace
from wx_speaker_control import SpeakerControlLocal
from wx_grid_path import GridPath
from braille_cell import BrailleCell
from magnify_info import MagnifyInfo, MagnifyDisplayRegion
from wx_stuff import *
from wx_adw_front_end import AdwFrontEnd
from wx_adw_menus import AdwMenus
from wx_canvas_panel import CanvasPanel
from wx_braille_cell_list import BrailleCellList
from wx_tk_rpc_user import TkRPCUser
from wx_canvas_panel_item import CanvasPanelItem

class AudioDrawWindow(wx.Frame):
    def __init__(self,
        tkr=None,
        snapshot_num=None,
        display_list=None,
        app=None,
        id_title = "unknown",
        title=None, speaker_control=None,
        win_width=800, win_height=800,
        grid_width=40, grid_height=25,
        win_fract=True,
        x_min=None, y_min=None,
        x_max=None, y_max=None,
        line_width=1, color="black",
        pos_check_interval= .1,
        pos_rep_interval = .1,
        pos_rep_queue_max = 1,
        visible = True,
        enable_mouse = False,
        pgmExit=None,
        blank_char=",",
        drawing=False,
        show_marked=False,
        shift_to_edge=None,
        silent=False,
        look_dist=2,
        menu_str="",
        key_str="",
        mag_info=None,
        setup_wx_win = True,
        iy0_is_top=True,        # OBSOLETE
                 ):
        self.id_title = id_title
        SlTrace.lg(f"""\nAudioDrawWindow:
        tkr={tkr},
        snapshot_num={snapshot_num},
        display_list={display_list},
        app={app},
        id_title={id_title},
        title={title},
        speaker_control={speaker_control},
        win_width={win_width}, win_height={win_height},
        grid_width={grid_width}, grid_height={grid_height},
        win_fract={win_fract},
        x_min={x_min}, y_min={y_min},
        x_max={x_max}, y_max={y_max},
        line_width={line_width}, color={color},
        pos_check_interval= {pos_check_interval},
        pos_rep_interval = {pos_rep_interval},
        pos_rep_queue_max = {pos_rep_queue_max},
        visible = {visible},
        enable_mouse = {enable_mouse},
        pgmExit={pgmExit},
        blank_char={blank_char},
        drawing={drawing},
        show_marked={show_marked},
        shift_to_edge={shift_to_edge},
        silent={silent},
        look_dist={look_dist},
        menu_str={menu_str},
        key_str={key_str},
        mag_info={mag_info},
        setup_wx_win = {setup_wx_win},
        iy0_is_top={iy0_is_top},        # OBSOLETE           
                   """, "adw")
        SlTrace.lg(f"AudioDrawWindow: {id_title = }")
        #frame = CanvasFrame(title=mytitle, size=wx.Size(width,height))
        """ Setup audio window
        :tkr: Access to remote tk information default: simulated access
        :make_snapshot: make a snapshot of current canvas
                        default: False - use current window state 
        :app: wx application object
            default: create object
        :speaker_control: (SpeakerControlLocal) local access to centralized speech making
        :win_width: display window width in pixels
            default: 800
        :win_height: display window height in pixels
            default: 800
        :win_fract: True - x_min,x_max are fraction of self.draw_width()
                            y_min,y_max are fractions of self.draw_height()
                    False - values
                    default: True
        :x_min: minimum coordinate
                default: win_fract: True: 0
                                False: 0
        :y_min: minimum coordinate
                default: win_fract: True:  0
                                    False: 0

        :x_max: maximum coordinate
                default: win_fract: True: 1
                                    False: self.draw_width()
        :y_max: maximum coordinate
                default: win_fract: True: 1
                                    False: self.draw_height()
        :grid_width: braille width in cells
            default: 40
        :grid_height: braille width in cells
            default: 25
        :title: window title
        :pos_check_interval: time between checks/reporting on
                cursor position  default: .1 seconds
        :pos_rep_interval: minimum time between reports
                default: .5 seconds
        :pos_rep_force_output: force output
                default: Force
        :pos_rep_queue_max: maximum position report queue maximum
                default: 4
        :visible: cells are visible
                default: True - visible
        :blank_char: replacement for non-trailing blanks
                    default "," to provide a "mostly blank",
                    non-compressed blank character for the
                    braille graphics
                    default: "," dot 2.
        :shift_to_edge: shift picture to edge/top
                    to aid in finding
                    default: False - no shift
        :drawing: We're in drawing mode
                 default: False - no drawing
        :show_marked: Make marked cells
                shown even when invisible
                default: True
        :silent: Suppress talking / audio
                default: False
        :menu_str:  Initial menu commands
                    default=None
                    Note that interpretation is case insensitive
                    Format: <Menu Letter><option letters>; as separator
                            e.g. n:sh;d:d  for Navigator silent, help
        :enable_mouse: mouse cursor operation enabled
                    default: False                                        Drawing: start drawing
        :key_str:  Initial key symbol command string
                    Each command separated by ";" to facilitate
                    recognition of multi charater symbols
                    such as "up", "down" Note that
                    interpretation is case insensitive
        :look_dist: Max number of cells to look ahead and report
                    default: 1
        :iy0_is_top: OBSOLETE
        """
        # direction for digit pad
        if app is None:
            app = wx.App()
        self.app = app
        if win_width is None:
            win_width = 800
        if win_height is None:
            win_height = 800
        super().__init__(None, title=title,
                         size=wx.Size(win_width, win_height))
        self.cells = {}         # Dictionary of cells by (ix,iy)
        self.cells_comps = {}   # Dictionary of cell composite items by (ix,iy)
        self._cursor_xy = None  # Cursor, if one, position
                                #   xy pair on canvas (0-max)
        self._cursor_rect = None    # Refresh rectangle, if one
        self._cursor_item = None    # position cursor tag
        self.snapshot_num = snapshot_num
        if tkr is None:
            tkr = TkRPCUser(simulated=True)
        self.tkr = tkr
        SlTrace.lg("USER: Linking AudioDrawWindow to tkr", "adw")
        tkr.set_adw(self)   # Support access, only first

        self.display_list = display_list
        if title is None:
            title = "AudioDrawWindow"
        self.title = title
        control_prefix = "AudioDraw"
        self.win_width = win_width
        self.win_height = win_height
        self.pgmExit = pgmExit
        if x_min is None:
            x_min = 0. if win_fract else 0.
        if y_min is None:
            y_min = 0. if win_fract else 0.
        if x_max is None:
            x_max = 1. if win_fract else self.draw_width()
        if y_max is None:
            y_max = 1. if win_fract else self.draw_height()
        # create the audio feedback window
        self.title = title
        
        self.Show()
        #self.adw_panel = wx.Panel(self)
        if speaker_control is None and setup_wx_win:
            SlTrace.lg("Creating own SpeakerControl")
            speaker_control = SpeakerControlLocal()
        self.speaker_control = speaker_control
        #mw.withdraw()
        
        ###wxport### For print entry support
        self.win_print_entry = None


        self._visible = visible
        self.grid_width = grid_width
        self.grid_height = grid_height
        self.canv_pan = CanvasPanel(self)
        if shift_to_edge is None:
            shift_to_edge = True
        self.shift_to_edge = shift_to_edge
        self.fte = AdwFrontEnd(self, title=title, silent=silent, color=color)
        self.canv_pan.set_key_press_proc(self.fte.key_press)
        self.menus = AdwMenus(self.fte, frame=self)
            
        self.cell_height = self.draw_height()/self.grid_height
        self.set_win_fract(win_fract)
        self.set_x_min(x_min)
        self.set_y_min(y_min)
        self.set_x_max(x_max)
        self.set_y_max(y_max)
        self.set_drawing(drawing)
        ###self.speak_text(title)

        self.escape_pressed = False # True -> interrupt/flush
        #self.set_cell_lims()
        self.do_talking = True      # Enable talking
        self.logging_speech = True  # Output speech to log/screen
        self.from_initial_canvas = False    # True iff from initial drawing
        nrows = grid_height
        ncols = grid_width
        if mag_info is None:
            self.from_initial_canvas = True     # Setup from original canvas
            top_region = MagnifyDisplayRegion(
                win_fract=win_fract,
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max,
                nrows=nrows,
                ncols=ncols)
            if not win_fract:
                x_min, x_max, y_min, y_max = tkr.get_canvas_lims()  # get actual limits
            display_region = MagnifyDisplayRegion(
                win_fract=win_fract,
                x_min=x_min,
                y_min=y_min,
                x_max=x_max,
                y_max=y_max,
                nrows=nrows,
                ncols=ncols)
            mag_info = MagnifyInfo(top_region=top_region,
                                   display_region=display_region)
            self.cell_specs = self.get_cell_specs(
                                        snapshot_num=self.snapshot_num,
                                        win_fract=win_fract,
                                        x_min=x_min,x_max=x_max,
                                        y_min=y_min,y_max=y_max)
        else:
            if mag_info.description:
                title = mag_info.description
                ###wxport###mw.title(title)
                self.title = title
                ###self.speak_text(f"Magnification of {title}")
        mag_info.display_window = self    # Magnification setup
        if mag_info.parent_info is not None:
            mag_info.parent_info.child_infos.append(mag_info)
        self.mag_info = mag_info
        SlTrace.lg(f"\nAudioDrawWindow() mag_info:\n{mag_info}\n",
                   "audio_draw_window")
        self.mag_selection_id = None       # mag selection canvas tag, if one
        self.is_selected = False            # Flag as not yet selected
        self.pos_history = []       # position history (ix,iy)
        self.pos_rep_force_output = False
        self.goto_travel_list = []    # Memory of where we have gone
        self.goto_travel_list_index = None
        self.cell_history = []          # History of all movement
        self.set_grid_path()
        self.running = True         # Set False to stop
        ###wxport###self.mw.focus_force()
        self.blank_char = blank_char
        self.set_look_dist(look_dist)
        #self.pos_check()            # Startup possition check loop
        self.fte.do_complete(menu_str=menu_str, key_str=key_str)
        self.setup_cells()
        self.update(full=True)     # Make visible
        self.Raise()                # put on top

    def setup_cells(self):
        """ Setup/display cells if any
        """
        if self.display_list is not None:
            display_list = self.display_list
            if type(display_list) == str:
                display_list = BrailleCellList().get_from_string(display_list)
            display_cells = {}
            for dc in display_list :
                ix,iy,color = dc.ix,dc.iy,dc._color
                dcell = BrailleCell(ix=ix, iy=iy,
                                    color=color)
                display_cells[(ix,iy)] = dcell
            if SlTrace.trace("cell_spec"):
                SlTrace.lg("from display_list - draw_cells")
        if self.from_initial_canvas:
            SlTrace.lg("from_initial_canvas - draw_cells", "adw")
            self.draw_cells(self.cell_specs)
        self.key_goto()      # Might as well go to figure
        self.find_edges()
        if True:
            self.get_cell_bounds("setup_cells")  # Show figure bounds

    def exit(self, rc=None):
        """ Main exit if creating magnifications
        """
        if rc is None:
            rc = 0
        if self.pgmExit is not None:
            self.pgmExit()      # Use supplied pgmExit
            
        SlTrace.lg("AudoDrawWindow.exit", "adw")
        SlTrace.onexit()    # Force logging quit
        os._exit(0)

    def set_look_dist(self, look_dist):
        self.look_dist = look_dist

    def get_look_dist(self):
        return self.look_dist

    def silence(self):
        """ Check if silent mode
        """
        return self.fte.silence()

    def speak_text(self, msg, dup_stdout=True,
                   msg_type=None,
                   rate=None, volume=None):
        """ Speak text, if possible else write to stdout
        :msg: text message, iff speech
        :dup_stdout: duplicate to stdout default: True
        :msg_type: type of speech default: 'REPORT'
            REPORT - standard reporting
            CMD    - command
            ECHO - echo user input
        :rate: speech rate words per minute
                default: 240
        :volume: volume default: .9            
        """
        self.win_print(msg)
        if self.is_silent():
            if dup_stdout:
                SlTrace.lg(msg)
            return
        
        self.speaker_control.speak_text(msg=msg, msg_type=msg_type,
                             dup_stdout=dup_stdout,
                             rate=rate, volume=volume)

    def stop_speak_text(self):
        """ Stop ongoing speach, flushing queue
        """
        self.speaker_control.stop_speak_text()
        
    def trav_len(self, ix, iy, dir_x, dir_y, require_cell=True):
        """ Find the travel length
        This is the number of steps from cell(ix,iy) in 
        direction (dir_x,dir_y) while still in self.cells[]
        traversal is stopped at traversed cells - in
        self.goto_cell_list
        :ix: cell x-index
        :iy: cell y_index
        :dir_x: x change each step
        :dir_y: y change each step
        :require_cell: Require cell in traversal default: True - required
        :returns: steps_forward, steps_backward, -1,-1
                 if cell, itself is not in figure(self.cells)
        """
        if require_cell and (ix,iy) not in self.cells:
            return -1,-1       # Not in figure
        
        if dir_x == 0 and dir_y == 0:
            return 0,0
        
        tlen_forward = 0
        ix_f = ix
        iy_f = iy
        while True:
            ix_f += dir_x
            iy_f += dir_y
            if (ix_f,iy_f) not in self.cells:
                break
            if (ix_f,iy_f) in self.get_goto_cell_list():
                break
            tlen_forward += 1   # traversal extended

        # Get backend of line
        tlen_backward = 0
        ix_b = ix
        iy_b = iy
        
        while True:
            ix_b -= dir_x
            iy_b -= dir_y
            if (ix_b,iy_b) not in self.cells:
                break
            if (ix_b,iy_b) in self.get_goto_cell_list():
                break
            tlen_backward += 1   # traversal extended
        return tlen_forward,tlen_backward
             
    def draw_cells(self, cells=None, show_points=False):
        """ Display braille cells on canvas
        :cells: list of BrailleCell or (ix,iy,color)tuple
                or dictionary of BrailleCell  cells
                to draw
        :show_points: instead of braille, show sample points
        """
        SlTrace.lg("draw_cells", "draw_cells")
        if type(cells) == str:
            tbstk = traceback.extract_stack()
            tbstk_lst = traceback.format_list(tbstk)
            tbstr = "\n".join(tbstk_lst)
            SlTrace.lg(f"draw_cells: {cells} - error?\n{tbstr}\n")
            return
        
        SlTrace.lg(f"x_min:{self.get_x_min()} y_min: {self.get_y_min()}", "draw_cells")
        if cells is None:
            cells = self.get_cells()
        else:
            if not isinstance(cells, dict):
                cs = {}
                for cell in cells:
                    bcell = BrailleCell.tuple_to_braille_cell(cell)
                    cs[(bcell.ix,bcell.iy)] = bcell
                cells = cs
            self.cells = cells      # Copy
        min_x, max_y, max_x,min_y = self.bounding_box(cells=cells)
        if min_x is not None:
            SlTrace.lg(f"Drawn cells bounding box", "draw_cells")            
            SlTrace.lg(f"Upper left: min_x:{min_x} max_y:{max_y}", "draw_cells")
            SlTrace.lg(f"Lower Right: max_x:{max_x} min_y:{min_y}", "draw_cells")
            self.get_cell_bounds("draw_cells", cells=cells)
        SlTrace.lg(f"{len(cells)} cells", "draw_cells")
        for cell in cells.values():
            self.display_cell(cell)
            cell.mtype = cell.MARK_UNMARKED
        if min_x is not None:
            self.set_xy((min_x,min_y))
            x,y = self.get_xy()
            #self.pos_check(x=x,  y=y)
        self.set_grid_path()
        self.pos_history = []       # Clear history
        self.set_scanning()         # Setup positioning info
                    
    def print_braille(self, title=None, shift_to_edge=None):
        """ Output braille display
            with an "identificatin title" created
            from source file name, date, login name
            to start each print for identification
        :title: title default: self.title
        :shift_to_edge: shift figure towards edge to ease finding figure
                        default: self.shift_to_edge
        """
        if self.id_title is not None:
            id_title = self.id_title
        else:
            id_title = "???"
        SlTrace.lg(f"print_braille: {id_title = }")
        if title is None:
            title = self.title
        if title is not None:
            print(title)
        if shift_to_edge is None:
            shift_to_edge = self.shift_to_edge
        if shift_to_edge:
            self.find_edges()
            left_edge = self.left_edge
            right_edge = self.right_edge
            top_edge = self.top_edge
            bottom_edge = self.bottom_edge
        else:
            left_edge = 0
            top_edge = 0
            right_edge = self.grid_width-1
            bottom_edge = self.grid_height-1

        braille_text = ""
        for iy in range(top_edge, bottom_edge):
            line = ""
            for ix in range(left_edge, right_edge+1):
                cell_ixy = (ix,iy)
                if cell_ixy in self.cells:
                    cell = self.cells[cell_ixy]
                    color = cell.color_string()
                    line += color[0]
                else:
                    line += " "
            line = line.rstrip()
            if self.blank_char != " ":
                line = line.replace(" ", self.blank_char)
            ###print(f"{iy:2}", end=":")
            braille_text += line + "\n"
        data = f"\n{id_title}"
        data += f"\n{title}\n"
        data += braille_text
        SlTrace.lg(f"clipboard data:\n{data}")
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(wx.TextDataObject(data))
            wx.TheClipboard.Close()
        else:
            SlTrace.lg("Can't paste to clipboard")

    def print_cells(self, title=None):
        """ Display current braille in a window
        """
        if title is not None:
            print(title)
        for ix in range(self.grid_width):
            for iy in range(self.grid_height):
                cell_ixy = (ix,iy)
                if cell_ixy in self.cells:
                    cell = self.cells[cell_ixy]
                    SlTrace.lg(f"ix:{ix} iy:{iy} {cell_ixy}"
                               f" {cell._color}"
                          f" rect: {self.get_cell_rect_tur(ix,iy)}"
                          f"  win rect: {self.get_cell_rect_win(ix,iy)}")
        SlTrace.lg("")

    def find_edges(self):
        """Find  top, left, bottom, right non-blank edges
        so we can shift picture to left,top for easier
        recognition
        :returns: left_edge, top_edge, right_edge, bottom_edge
                    Also sets self.left_edge,...
        """
        left_edge,top_edge, right_edge,bottom_edge = self.bounding_box_ci()
        if bottom_edge is None:
            bottom_edge = self.grid_height-1
                    
        if left_edge > 0:           # Give some space
            left_edge -= 1
        if left_edge > 0:
            left_edge -= 1
        self.left_edge = left_edge
        
        if top_edge > 0:           # Give some space
            top_edge -= 1
        if bottom_edge  < self.grid_height-1:
            bottom_edge += 1        
        self.top_edge = top_edge
        self.bottom_edge = bottom_edge
        
        self.left_edge = left_edge
        self.right_edge = right_edge
         
        return left_edge, top_edge, right_edge, bottom_edge

        
    def set_cursor_pos_win(self, x=0, y=0, quiet=False):
        """ Set mouse cursor position in win(canvas) coordinates
        :x: x-coordinate (win(canvas))
        :y: y-coordinate (win(canvas)) 
        :quiet: Don't announce legal moves
                default:False
        """
        """ Set mouse cursor position in canvas coordinates
        :x: x-coordinate (canvas win)
        :y: y-coordinate (canvas win)
        :quiet: Don't announce legal moves
                default: False
        """
        if not self.running:
            return
        
        self.set_xy((x,y))      # All we do is set coordinate memory
        loc_ixiy = self.get_ixy_at()
        self.add_to_pos_history(loc_ixiy)
        cell_ixiy = self.get_cell_at()
        if cell_ixiy is not None:
            self.cell_history.append(cell_ixiy)
            if not self.is_drawing():   # If we're not drawing
                self.mark_cell(cell_ixiy)   # Mark cell if one
        
        self.cursor_update()

        ###wxport###if not self.mw.winfo_exists():
        ###wxport###    return 
        
        #self.update()
        if not quiet:
            self.pos_check(force_output=True)

    def move_to_ixy(self, ix=None, iy=None):
        """ Move to grid (cell) ix,iy
        :ix: ix index default: current ix
        :iy: iy index default: current iy
        """
        ix_cur,iy_cur = self.get_ixy_at()
        if ix is None:
            ix = ix_cur 
        if iy is None:
            iy = iy_cur
        win_xc,win_yc = self.get_cell_center_win(ix,iy)
        self.move_to(win_xc,win_yc)



    def move_to(self, x,y, quiet=False):
        """ Move to window loc
        Stop at edges, with message
        :x:  win x-coordinate
        :y: win y-coordinate
        :quiet: Don't announce legal move
                default:False
        """
        self.set_cursor_pos_win(x=x, y=y, quiet=quiet)
                
    def get_ix_min(self):
        """ get minimum ix on grid
        :returns: min ix
        """
        return 0

    def get_ix_max(self):
        """ get maximum ix on grid
        :returns: max ix
        """
        return self.grid_width-1

    def get_iy_min(self):
        """ get minimum iy on grid
        :returns: min iy
        """
        return 0

    def get_iy_max(self):
        """ get maximum ix on grid
        :returns: min ix
        """
        return self.grid_height-1

    def get_mag_info(self):
        """ Get magnification info storage
        """
        return self.mag_info

    def get_speaker_control(self):
        """ Get speech control
        """
        return self.speaker_control

        

    def bounding_box(self, cells=None, add_edge=None):
        """ canvas coordinates which bound displayed figure
        :cells: list of cells, (with cell.ix,cell.iy) or (ix,iy) tuples
                default: list of all cells in figure
        :add_edge: number of cells to add/subtract (if possible)
                     to enlarge/shrink box
                    default: no change
        :returns: min_x, max_y, max_x, min_y  (upper left) (lower right)
                    None,None,None,None if no figure
        """
        min_ix, min_iy, max_ix, max_iy = self.bounding_box_ci(
                                    cells=cells, add_edge=add_edge)
        if min_ix is None:
            return None,None,None,None      # No figure
        
        min_x, max_y, _, _ = self.get_cell_rect_tur(min_ix,min_iy)
        _, _, max_x, min_y = self.get_cell_rect_tur(max_ix,max_iy)
        SlTrace.lg(f"bounding_box: min_x,max_y, max_x, min_y {min_x,max_y, max_x, min_y}",
                   "bounding_box")
        return min_x,max_y, max_x, min_y
    
    def bounding_box_ci(self, cells=None, add_edge=None):
        """ cell indexes which bound the list of cells
        :cells: list of cells, (with cell.ix,cell.iy) or (ix,iy) tuples
                default: list of all cells in figure
        :add_edge: number of cells to add/subtract (if possible)
                     to enlarge/shrink box
                    default: no change
        :returns: 
                    None,None,None,None if no figure
                    upper left ix,iy  lower right ix,iy
        """
        if cells is None:
            if not hasattr(self, "cells"):
                return None,None,None,None         # Not yet setup
            cells = list(self.cells.keys())
        
        ix_min, iy_max, ix_max,iy_min = None,None,None,None
        for cell in cells:
            cell_ixy = cell if isinstance(cell, tuple) else (cell.ix,cell.iy)
            cell_ix, cell_iy = cell_ixy
            if ix_min is None or cell_ix < ix_min:
                ix_min = cell_ix
            if ix_max is None or cell_ix > ix_max:
                ix_max = cell_ix
            if iy_min is None or cell_iy < iy_min:
                iy_min = cell_iy
            if iy_max is None or cell_iy > iy_max:
                iy_max = cell_iy

        if ix_min is None:
            ix_min = 0
        if ix_max is None:
            ix_max = self.grid_width-1
        if iy_min is None:
            iy_min = 0
        if iy_max is None:
            iy_max = self.grid_height-1
        
        if add_edge is not None:    # Extend/Shrink box
            ext_ix_min = ix_min - add_edge
            bd_ix_max = self.get_ix_max() # protect against too large of negative add_edge
            if ext_ix_min > bd_ix_max: ext_ix_min = bd_ix_max
            ix_min = max(ext_ix_min, self.get_ix_min()) # limit to bounds

            ext_iy_min = iy_min - add_edge
            bd_iy_max = self.get_iy_max()
            if ext_iy_min > bd_iy_max: ext_iy_min = bd_iy_max
            iy_min = max(ext_iy_min, self.get_iy_min())
            
            ext_ix_max = ix_max + add_edge
            bd_ix_min = self.get_ix_min()
            if ext_ix_max < bd_ix_min: ext_ix_max = bd_ix_min
            ix_max = min(ext_ix_max, self.get_ix_max())
            
            ext_iy_max = iy_max + add_edge
            bd_iy_min = self.get_iy_min()
            if ext_iy_max < bd_iy_min: ext_ix_max = bd_iy_min
            iy_max = min(ext_iy_max, self.get_iy_max())
        SlTrace.lg(f"bounding_box_ci: ix_min,iy_min, ix_max,iy_max: {ix_min,iy_min, ix_max,iy_max}",
                   "adw")    
        return ix_min,iy_min, ix_max,iy_max
            


    def erase_cell(self, cell, force_refresh=True):
        """ Erase cell
        :cell: BrailleCell
        """
        if cell is None:
            return
        
        # Remove current items, if any
        if cell.canv_items:
            for item_id in cell.canv_items:
                self.canv_pan.delete(item_id)
        cell.canv_items = []
        if force_refresh:
            self.refresh_cell(cell)      # Mark as dirty

    def erase_pos_history(self):
        """ Remove history, undo history marking
        """
        for cell_ixy in self.pos_history:
            if cell_ixy in self.cells:
                cell = self.cells[cell_ixy]
                self.mark_cell(cell, BrailleCell.MARK_UNMARKED)
                self.display_cell(cell)
        self.pos_history = []
        self.remove_mag_selection()    
        ###self.update()
                
    def display_reposition_hack(self, cx1,cx2,cy1,cy2, force=False):
        """ ###TFD HACK to reposition cell dispaly
            move x1,x2,y1,y2
        """
        return cx1,cx2,cy1,cy2

    def annotate_cell(self, cell_ixy=None, color=None,
                      outline="blue", outline_width=2,
                      text=None):
        """ Annotate cell to highlight it
        Possibly for perimeter viewing
        :cell_xy: ix,iy tuple default: current location
        :color: rectangle color default: no fill
        :outline: add outline color
                    default: no special outline
        :outline_width: outline width
                    default: 2
        :text: added text default: no text added
        """
        if cell_ixy is None:
            cell_ixy = self.get_ixy_at()
        cell = self.get_cell_at_ixy(cell_ixy=cell_ixy)
        if cell is None:
            return     # Just ignore if missing
        
        self.deannotate_cell(cell_ixy=cell_ixy) #???
        self.display_cell(cell=cell)
        canvas = self.canvas
        cx1,cy1,cx2,cy2 = self.get_win_ullr_at_ixy_canvas(cell_ixy)
        if outline is not None:
            canv_item = canvas.create_rectangle(cx1,cy1,cx2,cy2,
                                fill=color,
                                 outline=outline,
                                 width=outline_width)
            cell.canv_items.append(canv_item)
            
        self.update()    # So we can see it now 
            
        return 

    def deannotate_cell(self, cell_ixy=None, color=None,
                      outline=None, text=None):
        """ deAnnotate cell, restoring cell to normal
        :cell_xy: ix,iy tuple default: current location
        """
        if cell_ixy is None:
            cell_ixy = self.get_ixy_at()
        cell = self.get_cell_at_ixy(cell_ixy=cell_ixy)
        if cell is not None:
            self.display_cell(cell=cell)
            
        return      # Just ignore if missing

    def update_cell(self, cell, **kwargs):
        """ Update cell, with "inplace" attribute changes 
        :**kwargs: attributes to be changed
        """
        items = self.get_cell_items(cell)   # Get rectangle item(s)
        if len(items) > 0:
            self.canv_pan.update_item(items[0], **kwargs)
            self.canv_pan.refresh_item(items[0])
        
        
    def display_cell(self, cell, show_points=False):
        """ Display cell
        :cell: BrailleCell
        :show_points: show points instead of braille
                default: False --> show braille dots
        """
        if not cell.is_visible():
            self.erase_cell(cell)
            return              # Nothing to show
        
        ix = cell.ix
        iy = cell.iy
        comp_id = self.canv_pan.create_composite(disp_type=CanvasPanelItem.DT_CELL,
                                                 desc=str(cell))
        cell.canv_items.append(comp_id)
        comp_item = self.canv_pan.id_to_item(comp_id)
        cell.comp_item = comp_item
        self.canv_pan.add_cell(comp_item)   
        cx1,cy1,cx2,cy2 = self.get_win_ullr_at_ixy_canvas((ix,iy))
        SlTrace.lg(f"{ix},{iy}: {cell} :{cx1},{cy1}, {cx2},{cy2} ", "display_cell")
        
        if cell.mtype==BrailleCell.MARK_UNMARKED:
            canv_id = self.canv_pan.create_rectangle(cx1,cy1,cx2,cy2,
                                    fill="#d3d3d3",
                                    outline="dark gray")
        else:
            canv_id = self.canv_pan.create_rectangle(cx1,cy1,cx2,cy2,
                                    fill="#b0b0b0",
                                    outline="dark gray")
        comp_item.add(canv_id)
        
        color = self.color_str(cell._color)
        if len(color) < 1 or color[0] not in BrailleCell.color_for_character:
            color = "black"
        if show_points:
            dot_size = 1            # Display cell points
            dot_radius = dot_size//2
            if dot_radius < 1:
                dot_radius = 1
                dot_size = 2
            for pt in cell.points:
                dx,dy = self.get_point_win(pt)
                x0 = dx-dot_radius
                y0 = dy+dot_size 
                x1 = dx+dot_radius 
                y1 = dy
                canv_id = self.canv_pan.create_oval(x0,y0,x1,y1,
                                                fill=color)
                cell.canv_items.append(canv_id)
                SlTrace.lg(f"canv_pan.create_oval({x0},{y0},{x1},{y1}, fill={color})", "aud_create")
        else:            
            dots = cell.dots
            cell_width = cx2-cx1
            cell_height = cy2-cy1       # y increases down
            # Fractional offsets from lower left corner
            # of cell rectangle
            ll_x = cx1      # Lower left corner
            ll_y = cy1
            ox1 = ox2 = ox3 = .3 
            ox4 = ox5 = ox6 = .7
            oy1 = oy4 = .15
            oy2 = oy5 = .45
            oy3 = oy6 = .73
            dot_size = .25*cell_width   # dot size fraction
            dot_radius = dot_size//2
            dot_offset = {1: (ox1,oy1), 4: (ox4,oy4),
                        2: (ox2,oy2), 5: (ox5,oy5),
                        3: (ox3,oy3), 6: (ox6,oy6),
                        }
            for dot in dots:
                offsets = dot_offset[dot]
                off_x_f, off_y_f = offsets
                dx = ll_x + off_x_f*cell_width
                dy = ll_y + off_y_f*cell_height
                x0 = dx-dot_radius
                y0 = dy+dot_size 
                x1 = dx+dot_radius 
                y1 = dy
                canv_id = self.canv_pan.create_oval(x0,y0,x1,y1,
                                                fill=color)
                comp_item.add(canv_id) 
                SlTrace.lg(f"canv_pan.create_oval({x0},{y0},{x1},{y1}, fill={color})", "aud_create")
        self.refresh_cell(cell) 
        self.cursor_update()
        
    def display_cell_end(self, cell):
        """ Complete cell display
        :cell: BrailleCell to display
        """
        self.refresh_cell(cell) 


    def cell_mtype_to_fill(self, mtype=BrailleCell.MARK_SELECTED):
        """ Convert cell mtype to fill color
        :mtype: cell mark type default: selected
        :returns: fill string
        """
        MARK_UNSELECTED = 1
        MARK_SELECTED = 2
        MARK_TRAVERSED = 3

        fill = "#b0b0b0"    # 
        if mtype == BrailleCell.MARK_SELECTED:
            fill = "#d3d3d3"
        elif mtype == BrailleCell.BrailleCell.MARK_UNSELECTED:
            fill = "#b0b0b0"
        return fill

    def mark_cell(self, cell,
                  mtype=BrailleCell.MARK_SELECTED):
        """ Mark cell for viewing of history
        :cell: Cell(BrailleCell) or (ix,iy) of cell
        :mtype: Mark type default: MARK_SELECTED
        """
        if isinstance(cell, BrailleCell):
            ixy = (cell.ix,cell.iy)
        else:
            ixy = cell
        cells = self.get_cells()
        if ixy in cells:
            cell = cells[ixy]
            cell.mtype = mtype
            #fill = self.cell_mtype_to_fill(mtype)
            #self.update_cell(cell, fill=fill)
            self.display_cell(cell)
            self.refresh_cell(cell)

    def draw_height(self):
        """ drawing area height reduced from window height
        """
        return self.win_height*(21/25)  # Hack for now

    def draw_width(self):
        """ drawing area width reduced from window width
        """
        return self.win_width*(38/40)  # Hack for now
    
    def partial_update_add(self, item):
        """ Add item to partial update
        :item: item to add
        """
        self.canv_pan.partial_update_add(item=item)
        
    def partial_update_complete(self):
        """ Complete partial_update
        """
        self.canv_pan.partial_update_complete()
            
    def partial_update_is_in(self):
        """ Check if in partial update
        """
        return self.canv_pan.partial_update_is_in()
    
    def partial_update_start(self, rect=None):
        """ Setup a partial window update
        :rect: rectangle if present
        """
        self.canv_pan.partial_update_start(rect=rect)
        

                
    def get_cell_center_win(self, ix, iy):
        """ Get cell's window rectangle x, y  upper left, x,  y lower right
        :ix: cell x index
        :iy: cell's  y index
        :returns: window(0-max): (xc,yc) where
            xc,yc are x,y coordinates of cell's center
        """
        x1,y1, x2,y2 = self.get_cell_rect_win(ix,iy)
        return (x1+x2)/2, (y1+y2)/2 
                
    def get_cell_rect_win(self, ix, iy):
        """ Get cell's window rectangle x, y  upper left, x,  y lower right
        :ix: cell x index
        :iy: cell's  y index
        :returns: window(0-max): (x1,y1,x2,y2) where
            x1,y1 are lower left coordinates
            x2,y2 are upper right coordinates
        """
        win_x1,win_y1,win_x2,win_y2 = self.get_win_ullr_at_ixy((ix,iy))
        return (win_x1,win_y1,win_x2,win_y2)
        
                    
        
        
    def get_ixy_at(self, pt=None):
        """ Get cell(indexes) in which point resides
        If on an edge returns lower cell
        If on a corner returns lowest cell
        :pt: x,y pair location
                        in win coordinates
                default: current location
        :returns: ix,iy cell pair
                if out of bounds limit to min/max of ix,iy
        """
        if pt is None:
            pt = self.get_xy()
        x,y = pt
        x_min = self.get_x_min()
        y_min = self.get_y_min()
        ix = int((x-x_min)/self.draw_width()*self.grid_width)
        iy = int((y-y_min)/self.draw_height()*self.grid_height)
        #ix = int((x)/self.self.draw_height()*self.grid_width)        # TFD
        #iy = int((y)/self.self.draw_height()*self.grid_height)      # TFD
        return (ix,iy)

    def get_cells(self):
        """ Get cell dictionary (by (ix,iy)
        """
        return self.cells
    
    def iterate_cells(self):
        """ Get an iteratable to traverse cells (BrailleCell)
        :returns: iteratable through cells (BrailleCell)
        """
        return self.get_cells().values()

    def get_cell_rect(self, cell):
        """ Get cell's wx.Rect bounds
        :returns: bounding rectangle wx:Rect
        """
        ix = cell.ix
        iy = cell.iy    
        cx1,cy1,cx2,cy2 = self.get_win_ullr_at_ixy_canvas((ix,iy))
        rect = wx.Rect(wx_Point(cx1,cy1), wx_Point(cx2,cy2))
        return rect
        
    def get_cell_items(self, cell, canv_type="create_rectangle"):
        """ Get cell's item (rectangle)
        :cell: BrailleCell
        :returns: item (in canvas_panel), None if no item
        """
        ix = cell.ix
        iy = cell.iy    
        cx1,cy1,cx2,cy2 = self.get_win_ullr_at_ixy_canvas((ix,iy))
        loc = wx.Point(int((cx1+cx2)/2),
                       int((cy1+cy2)/2))
        items = self.canv_pan.get_panel_items(loc, canv_type=canv_type)
        return items

    def get_overlapping_items(self, item):
        """ Get display items overlapping item
        :item: item(CanvasPanelItem)/id
        :returns: list of overlapping items (CanvasPanelItem)
        """
        item_d = {}     # dictionary by (id)
        if type(item) == int and item in self.canv_pan.items_by_id:
            item = self.canv_pan.items_by_id[item]
        displayed_items = self.get_displayed_items()
        
        for displayed_item in displayed_items:
            if self.is_overlapping(item, displayed_item):
                did = displayed_item.get_id()
                item_d[did] = displayed_item
        item_cells = list(item_d.values())
        return item_cells
            
                   
    def get_cell_at(self, pt=None):
        """ Get cell at location, if one
        :pt: x,y pair location in turtle coordinates
                default: current location
        :returns: cell if at, else None
        """
        cell_ixy = self.get_ixy_at(pt)
        if cell_ixy is None:
            return None 
        
        if cell_ixy in self.cells:
            return self.cells[cell_ixy]
        
        return None
    
    def get_braille_cells(self, 
                        x_min=None, y_min=None,
                        x_max=None, y_max=None,
                        ncols=None, nrows=None):
        """ Get braille cells from remote tk canvas
        :returns: list of BrailleCell
        """        
        
        specs = self.get_cell_specs(
                        win_fract=self.win_fract,
                        x_min=x_min, y_min=y_min,
                        x_max=x_max, y_max=y_max,
                        n_cols=ncols, n_rows=nrows)

        braille_cells = []
        for spec in specs:
            ix,iy,color = spec
            bcell = BrailleCell(ix=ix, iy=iy, color=color)
            braille_cells.append(bcell)
        return braille_cells
    
    def get_cell_specs(self,
                       snapshot_num=None,
                       win_fract=True, 
                        x_min=None, y_min=None,
                        x_max=None, y_max=None,
                        n_cols=None, n_rows=None):
        """ Get cell specs from remote tk canvas
        :returns: list of cell specs (ix,iy,color)
        """
        if snapshot_num is None:
            snapshot_num = self.snapshot_num        
        cell_specs = self.tkr.get_cell_specs(
                        snapshot_num=snapshot_num,
                        win_fract=win_fract,
                        x_min=x_min, y_min=y_min,
                        x_max=x_max, y_max=y_max,
                        n_cols=n_cols, n_rows=n_rows)
        #SlTrace.lg(f"AudioDrawWindow.get_cell_specs: {cell_specs}")
        return cell_specs

    def get_canvas_lims(self, win_fract=True):
        """ Get canvas limits - internal values, to which
        self.base.find_overlapping(cx1,cy1,cx2,cy2) obeys.
        NOTE: These values, despite some vague documentation, may be negative
              to adhere to turtle coordinate settings.
        
        :returns: internal (xmin, xmax, ymin, ymax)
        """
        return self.tkr.get_canvas_lims(win_fract=win_fract)



    def set_grid_path(self):
        self.grid_path = GridPath(self)
        
    def get_grid_path(self):
        return self.grid_path
    
    def get_win_ullr_at_ixy(self, ixy):
        """ Get window rectangle for cell at ixy
        :ixy: cell index tupple (ix,iy)
        :returns: (x_left,y_top, x_right,y_bottom)
        """
        ix,iy = ixy
        x_min = self.get_x_min()
        y_min = self.get_y_min()
        x_left = int(ix*(self.draw_width()/self.grid_width) + x_min)
        x_right = int((ix+1)*(self.draw_width()/self.grid_width) + x_min)
        y_top = int((iy*(self.draw_height()/self.grid_height) + y_min))
        y_bottom = int((iy+1)*(self.draw_height()/self.grid_height) + y_min)
        return (x_left,y_top, x_right,y_bottom)

    def get_win_ullr_at_ixy_canvas(self, ixy):
        """ Get window rectangle for cell at ixy
        :ixy: cell index tupple (ix,iy)
        :returns: (x_left,y_top, x_right,y_bottom) canvas coords
        """
        x_l,y_t, x_r,y_b = self.get_win_ullr_at_ixy(ixy)
        x_left,y_top = self.win_to_canvas((x_l,y_t))
        x_right,y_bottom = self.win_to_canvas((x_r,y_b))
        return (x_left,y_top, x_right,y_bottom)
        
    def win_to_canvas(self, xy):
        """ Convert window coordinates to canvas (min to max) to (0,max)
        :xy: x,y tuple
        :returns: x,y canvas tuple (canv_panel)
        """
        x,y = xy
        canvas_x, canvas_y = (x-self.get_x_min(), y-self.get_y_min())
        return (canvas_x,canvas_y)
            
    def get_cell_at_ixy(self,cell_ixy, cells=None, ):
        """ Get cell at (ix,iy), if one
        :cell_ixy: (ix,iy)
        :cells: dictionary of cells by (ix,iy)
                default: self.get_cells()
        :returns: BrailleCell if one, else None
        """
        if cells is None:
            cells = self.get_cells()
        if cell_ixy in cells:
            return cells[cell_ixy]
        
        return None
        

    def is_at_cell(self, pt=None):
        """ Check if at cell
        :pt: x,y pair location in turtle coordinates
                default: current location
        :returns: True if at cell, else False
        """
        if self.get_cell_at(pt) is None:
            return False 
        
        return True

    def is_space(self, ixy, cells=None):
        """ Are we at a space
        :ixy: cell ix,iy indexes
        :cells: cells to check
                default: our cells - get_cells()
        :returns: True if a space (not a cell) 
        """
        if cells is None:
            cells = self.get_cells()
        cell = self.get_cell_at_ixy(cell_ixy=ixy, cells=cells)
        return True if cell is None else False

    def refresh_cell(self, cell):
        """ Mark cell dirty, in need of repainting
        :cell: BrailleCell
        """
        brect = self.get_cell_rect(cell)
        SlTrace.lg(f"refresh_cell:{(cell.ix,cell.iy)} {brect}", "refresh")
        self.canv_pan.refresh_rectangle(brect)
                
    def set_cell(self, pt=None, color=None):
        """ Set cell at pt, else current cell
        If no cell at current location create cell
        :pt: at point
            default: current location
        :color: color default: current color
        """
        if not self.is_at_cell(pt):
            self.create_cell(color=color)
        self.change_cell(pt=pt, color=color)

    def change_cell(self, pt=None, color=None):
        """ Change cell at pt if None current pt
        :pt: location default: current
        :color: color default: current color
        :returns: updated cell
        """
        if color is None:
            color = self.get_color()
        cell = self.get_cell_at(pt=pt)
        if cell is None:
            cell = self.create_cell(pt=pt, color=color)
        if cell is not None:
            cell.color_cell(color=color)
            self.display_cell(cell)
        return cell
                
    def color_str(self, color=None):
        """ Return color string
        :color: color specification str or tuple
        """
        color_str = color
        if (color_str is None
             or (isinstance(color_str, tuple)
                  and len(color_str) == 0)
             ):
            color_str = self.get_color()
        if isinstance(color_str,tuple):
            if len(color_str) == 1:
                color_str = color_str[0]
            else:
                color_str = "pink"  # TBD - color tuple work
        return color_str

    def is_visible(self):
        """ Check on visible mode
        """
        return self._visible
    
    def set_visible(self, val=True):
        """ Set cells visible/invisible
        Useful to give sighted a vision
        :val: set visible Default: True
        """
        SlTrace.lg(f"set_visible:{val}")
        self._visible = val
        if not val:
            self.set_show_marked(False)     # Make marked also invisible
        for cell in self.cells.values():
            self.set_visible_cell(cell, val)
                
    def set_visible_cell(self, cell, val=True):
        """ Set cells visible/invisible
        Useful to give sighted a vision
        :cell: figure cell
        :val: set visible Default: True
        """
        self.fte.set_visible_cell(cell, val=val)

    def announce_can_not_do(self, msg=None, val=None):
        """ Announce we can't do something
        """
        audio_beep = self.get_audio_beep()
        if audio_beep:
            audio_beep.announce_can_not_do(msg=msg, val=val)
        else:
            self.say_text(msg)
    
    def add_to_pos_history(self, loc_ixiy):
        """ Accumulate position history
        :loc_ixiy: (ix,iy) of current location
        """
        self.pos_history.append(loc_ixiy)   # location history
        SlTrace.lg(f"pos_history:{loc_ixiy}", "pos_tracking")

    
    def braille_for_color(self, color):
        """ Return dot list for color
        :color: color string or tuple
        :returns: list of dots 1,2,..6 for first
                letter of color
        """
        
        if color is None:
            color = self.get_color()
        if color is None:
            color = ("black")
        color = BrailleCell.color_str(color)
        c = color[0]
        dots = BrailleCell.braille_for_letter(c)
        return dots
    
    def braille_for_letter(self, c):
        """ convert letter to dot number seq
        :c: character
        :returns: dots tupple (1,2,3,4,5,6)
        """
        return BrailleCell.braille_for_letter(c)

    def clear_cells(self):
        """ remove all cells - clearing board
        Initialize empty board
        """
        SlTrace.lg("\nclear_cells")
        cells = self.get_cells()
        for cell in list(cells.values()):
            self.erase_cell(cell)
        self.draw_cells(cells=self.cells)
        del self.cells
        self.cells = {}
        self.clear()
        self.key_pendown(False) # Raise pen off paper
        
        
        
    def create_cell(self, cell_ixy=None, color=None,
                    show=True):
        """ Create new cell ad cell_xy
        :cell_xy: ix,iy tuple default: current location
        :color: color default: curren color
        :show: show cell default:True
        :returns: cell
        """
        if cell_ixy is None:
            cell_ixy = self.get_ixy_at()
        if color is None:
            color = self.get_color()
        dots = self.braille_for_color(color)
        bc = BrailleCell(ix=cell_ixy[0], iy=cell_ixy[1],
                         dots=dots, color=color)
        if cell_ixy in self.cells:
            del self.cells[cell_ixy]
        self.cells[cell_ixy] = bc
        if show:
            self.display_cell(bc)
        return bc
            
    def complete_cell(self, cell, color=None):
        """ create/Fill braille cell
            Currently just fill with color letter (ROYGBIV)
        :cell: (ix,iy) cell index or BrailleCell
        :color: cell color default: current color
        :returns: created/modified cell
        """
        if color is None:
            color = self.get_color()
        dots = self.braille_for_color(color)
        bc = BrailleCell(ix=cell[0],iy=cell[1], dots=dots, color=color)
        self.cells[cell] = bc
        return bc

    def update(self, x1=None, y1=None, x2=None, y2=None, full=False):
        """ Update display
            If x1,...y2 are present - limit update to rectangle
        """
        self.canv_pan.update(x1=x1, y1=y1, x2=x2, y2=y2, full=full)

    def MainLoop(self):
        self.app.MainLoop()
        
    def update_idle(self):
        """ Update pending
        """
        #self.Refresh()

    """ Creation/Modification Code
    """

    def create_audio_window(self, title=None,
                 snapshot_num=None,
                 win_width=None,
                 win_height=None,
                 win_fract=True,
                 xmin=None, xmax=None, ymin=None, ymax=None,
                 nrows=None, ncols=None, mag_info=None, pgmExit=None,
                 require_cells=False,
                 silent=False,
                 cell_specs=None):
        """ Create new AudioDrawWindow to navigate canvas from the section
        :title: optional title
                region (xmin,ymin, xmax,ymax) with nrows, ncols
        :snapshot_num: create a snapshot of current window
                default: track main window
        :win_width: window width default: self.win_width
        :win_height: window height default: self.draw_height()
        :speaker_control: (SpeakerControlLocal) local access to centralized speech facility
        :win_fract: True - xmin,xmax... are fractions 0. to 1. of window region
        :xmin,xmax,ymin,ymax,: see get_grid_lims()
                        default: CanvasGrid instance values
        :nrows,ncols: grid size for scanning
                default: if mag_info present: mag_info.mag_nrows, .mag_ncols
        :mag_info: magnification info (MagnifyInfo)
                    default: None
        :pgm_exit: function to call upon exit request
        :require_cells: Require at least some display cells to 
                    create window
                    default: False - allow empty picture
        :silent: quiet mode default: False
        :cell_specs: cells to display default: check for cells
        :returns: AudioDrawWindow instance or None if no cells
                Stores number of cells found in self.n_cells_created
        """
        if title is None:
            title = self.title
        self.title = title
        if pgmExit is None:
            pgmExit = self.exit 

        if mag_info is not None:
            title = f"Magnification {mag_info.description}"
        else:
            mag_info = self.create_magnify_info(
                                          x_min=xmin, y_min=ymin,
                                          x_max=xmax, y_max=ymax,
                                          ncols=ncols, nrows=nrows)
        if snapshot_num is None:
            snapshot_num = self.snapshot_num
                
        if nrows is None:
            nrows = mag_info.mag_nrows
        if ncols is None:
            ncols = mag_info.mag_ncols
        if cell_specs is None:        
            cell_specs = self.get_cell_specs(win_fract=win_fract,
                                            x_min=xmin, y_min=ymin,
                                            x_max=xmax, y_max=ymax,
                                            n_cols=ncols, n_rows=nrows)    
        self.n_cells_created = len(cell_specs)
        SlTrace.lg(f"create_audio_window:"
                   f" {self.n_cells_created} cells")
        SlTrace.lg(f"cell_specs: {cell_specs}", "cell_specs")
        if self.n_cells_created == 0:
            if require_cells:
                return None
                
        adw = AudioDrawWindow(tkr=self.tkr,
                              snapshot_num=snapshot_num,
                              app=self.app,
                              id_title=self.id_title,
                              title=title,
                              speaker_control=self.speaker_control,
                              iy0_is_top=True, mag_info=mag_info,
                              pgmExit=pgmExit,
                              win_width=win_width,
                              win_height=win_height,
                              x_min=xmin,
                              y_min=ymin,
                              x_max=xmax,
                              y_max=ymax,
                              silent=silent)
        adw.draw_cells(cells=cell_specs)
        adw.key_goto()      # Might as well go to figure
        return adw


    def create_magnify_info(self, x_min=None,y_min=None,
                    x_max=None,y_max=None,
                    ncols=None, nrows=None):
        """ Create a MagnifyInfo, using our values as defaults
        :x_min: minimum x value - left side 
        :y_min: minimum y value - top side
        :x_max: maximum x value - right side
        :y_max: maximum  y value - bottom side
        :ncols: number of grid columns
        :nrows: number of grid rows
        """
        if x_min is None:
            x_min = self.get_x_min()
        if y_min is None:
            y_min = self.get_y_min()
        if x_max is None:
            x_max = self.get_x_max()
        if y_max is None:
            y_max = self.get_y_max()
        if ncols is None:
            ncols = self.grid_width
        if nrows is None:
            nrows = self.grid_width
        
        top_region = MagnifyDisplayRegion(x_min=x_min, y_min=y_min,
                                          x_max=x_max, y_max=y_max,
                                          ncols=ncols, nrows=nrows)
        mag_info = MagnifyInfo(top_region=top_region,
                               base_canvas=self)
        return mag_info
                   
    def get_cell_bounds(self, title=None, cells=None, add_edge=0, display_region=None,
                        special=False):
        """ Get cell list bounds
        :title: optional title display default: no title
        :cells: list of cells, (with cell.ix,cell.iy) or (ix,iy) tuples
                default: list of all cells in figure
        :add_edge: number of cells to add/subtract (if possible)
                     to enlarge/shrink box
                    default: no change
        :display_region: (MagnifyDisplayRegion) display region
        :special: True - suppress extra display e.g. non-None cells processing
        :returns: xmin,ymin (upper left), xmax,ymax (lower right) display
        """
        if title is not None:
            SlTrace.lg(f"\n{title}", "cell_bounds")
        if not special and cells is not None:
            SlTrace.lg("Full display bounds", "cell_bounds")
            self.get_cell_bounds(cells=[(0,0), (self.grid_width-1,self.grid_height-1)],
                                 special=True)
            SlTrace.lg("\nFull figure bounds", "adw")
            self.get_cell_bounds(cells=None)
        ix_min, iy_min, ix_max, iy_max = self.bounding_box_ci(cells, add_edge)
        if display_region is None:
            di = self.get_mag_info()
            if di is None:
                SlTrace.lg("No Magnification Info")
                return None,None,None,None
            display_region = di.display_region
        dr = display_region
        disp_x_cell = (dr.x_max-dr.x_min)/dr.ncols
        disp_y_cell = (dr.y_max-dr.y_min)/dr.nrows
        xmin = ix_min*disp_x_cell + dr.x_min
        ymin = iy_min*disp_y_cell + dr.y_min
        xmax = (ix_max+1)*disp_x_cell + dr.x_min
        ymax = (iy_max+1)*disp_y_cell + dr.y_min
        SlTrace.lg(f"bounding indexes: ix_min:{ix_min}, iy_min:{iy_min}"
                   f", ix_max:{ix_max}, iy_max:{iy_max}", "cell_bounds")
        SlTrace.lg(f"cell bounds:"
                   f" xmin:{xmin} ymin:{ymin} xmax:{xmax} ymax:{ymax}"
                   f" nrows:{dr.nrows} ncols:{dr.ncols}", "cell_bounds")
                   
    def create_magnification_window(self, mag_info, adj_cell=.1):
        """ Create magnification
        :mag_info: MagnificationInfo containing info
        :adj_cell: Expansion adjustment fraction of cell
                default: 0
        :returns: instance of AudioDrawWinfow or None if none was created
        """
        select = mag_info.select
        disp_region = mag_info.display_region
        disp_x_cell = (disp_region.x_max-disp_region.x_min)/disp_region.ncols
        disp_y_cell = (disp_region.y_max-disp_region.y_min)/disp_region.nrows
        xmin = select.ix_min*disp_x_cell + disp_region.x_min - adj_cell*disp_x_cell
        ymin = select.iy_min*disp_y_cell + disp_region.y_min - adj_cell*disp_y_cell
        xmax = (select.ix_max+1)*disp_x_cell + disp_region.x_min + adj_cell*disp_x_cell
        ymax = (select.iy_max+1)*disp_y_cell + disp_region.y_min + adj_cell*disp_y_cell
        SlTrace.lg(f"create_magnification_window:"
                   f" xmin:{xmin} ymin:{ymin} xmax:{xmax} ymax:{ymax}"
                   f" nrows:{disp_region.nrows} ncols:{disp_region.ncols}", "adw")
        child_info = mag_info.make_child()
        child_info.display_region = MagnifyDisplayRegion(x_min=xmin, x_max=xmax,
                                        y_min=ymin, y_max=ymax)
        child_info.description = (f"{child_info.info_number}"
                                  + f" region min x: {xmin:.2f}, min y: {ymin:.2f},"
                                  + f" max x: {xmax:.2f}, max y: {ymax:.2f}")
        SlTrace.lg(f"create_audio_window: snapshot_num={self.snapshot_num}")
        adw = self.create_audio_window(snapshot_num=self.snapshot_num,
                                       xmin=xmin, xmax=xmax,
                                       ymin=ymin, ymax=ymax,
                                       nrows=disp_region.nrows,
                                       ncols=disp_region.ncols,
                                       mag_info=child_info,
                                       require_cells=True)            
        return adw 


    def cursor_update(self):
        """ Update cursor (current position) display
        Cursor is displayed at end of OnPaint
        
        """
        self._cursor_xy = self.get_xy_canvas()
        self.canv_pan.refresh_cursor()   # Remove old cursor, if any

        if self.mag_selection_id is not None:
            self.canv_pan.add_item(self.mag_selection_id)

    # REPLACED to set cursor
    # which is diplayed at end of OnPaint()
    def cursor_update_OLD(self):
        """ Update cursor (current position) display
        """
        self.remove_cursor()
        self.add_cursor()
        self.refresh_cursor()
        if self.mag_selection_id is not None:
            self.canv_pan.add_item(self.mag_selection_id)


    def refresh_cursor(self):
        """ Refresh cursor update
        """
        if self._cursor_item is not None:
            self.refresh_item(self._cursor_item)
                
    def remove_cursor(self):
        """ Remove old cursor if one"""
        if self._cursor_item is not None:
            overlapping_items = self.get_overlapping_items(self._cursor_item)
            self.canv_pan.delete(self._cursor_item)
            self._cursor_item = None
            for item in overlapping_items:
                self.canv_pan.add_item(item)
    
    def add_cursor(self):
        """ Add new cursor display
        """            
        rd = 5
        pos_x,pos_y = self.get_xy_canvas()

        x0 = pos_x-rd
        x1 = pos_x+rd
        y0 = pos_y-rd
        y1 = pos_y+rd
        self._cursor_item = self.canv_pan.create_cursor(x0,y0,x1,y1,
                                                    disp_type=CanvasPanelItem.DT_CURSOR,
                                                    fill="red")
        self.canv_pan.add_cursor(self._cursor_item)
        overlapping_items = self.get_overlapping_items(self._cursor_item)
        for item in overlapping_items:
                self.canv_pan.add_item(item)

    """
    Links to canvas_grid
    """
        
    def get_cell_rect_tur(self, ix, iy):
        """ Get cell's turtle rectangle x, y  upper left, x,  y lower right
        :ix: cell x index
        :iy: cell's  y index
        :returns: (min_x,max_y, max_x,min_y)
        """
        return self.tkr.get_cell_rect_tur(ix=ix, iy=iy)

    """
    Links to canvas panel - most are direct
    """
    
    def clear(self):
        """ Clear panel
        """
        self.canv_pan.clear()
        self.set_cursor_pos_win()
        
    def redraw(self):
        """ Redraw screen
        """
        self.canv_pan.redraw()

    def refresh_item(self, item):
        self.canv_pan.refresh_item(item)

    def get_displayed_items(self):
        """ Get list of displayed items
        :returns: list of permanently displayed values (AdwDisplayPendingItem)
        """
        return self.canv_pan.get_displayed_items()

    def is_overlapping(self, item1, item2):
        """ Check if two display items are overlapping
        :returns: True iff overlapping
        """
        return self.canv_pan.is_overlapping(item1, item2)

    """
         Links to front end functions
    """
    def is_pendown(self):
        """ Test if pen is down (marking)
        """
        return self.fte.is_pendown()
    
    
    def key_pendown(self, val=True):
        """ Lower/raise pen (starting,stopping) drawing
        :val: lower pen, if True else raise pen
        """
        self.fte.key_pendown(val)


    def set_initial_location(self):
        """ Set/Reset initial location of cursor
        """
        self.fte.set_initial_location()


    def is_inbounds(self,ix=None, iy=None):
        """ Test if inbounds
        :ix: x index - default: don't test
        :iy: y index - default: don't test
        :returns: True iff in bounds
        """
        return self.fte.is_inbounds(ix=ix, iy=iy)

    def remove_mag_selection(self):
        """ Remove magnify selection and marker
        """
        self.fte.remove_mag_selection()

    def set_show_marked(self,val=True):
        """ Show marked "invisible" cells
        """
        self.fte.set_show_marked(val=val)

    def wait_on_output(self):
        """ Wait till queued output speech/tones completed
        """
        self.fte.wait_on_output()
        
    def win_print(self,*args, dup_stdout=False, **kwargs):
        """ print to listing area
        :*args: print-like args
        :**kwargs: print-flags
        :dup_stdout:  send duplicate to stdout
        """
        self.fte.win_print(args=args, dup_stdout=dup_stdout, kwargs=kwargs)

    def get_audio_beep(self):
        """ Get access to audio beep
        """
        return self.fte.get_audio_beep()

    def set_color(self, color):
        """ Set current color
        """
        self.fte.set_color(color=color)
        
    def get_color(self):
        """ Get current color
        """
        return self.fte.get_color()
        
    def set_drawing(self, val=True):        
        self.fte.set_drawing(val=val)
        return val
    
    def is_drawing(self):
        return self.fte.is_drawing()

        
    def key_goto(self):
        """ Go to closest figure
            go to figure not in one already
            else go one step within current figure, toward
            longest inside path
        """
        self.fte.key_goto()

    def set_scanning(self, cells=None):
        self.fte.set_scanning(cells=cells)
        
    def get_vol(self, ix, iy, eye_ixy_l=None, eye_ixy_r=None):
        """ Get tone volume for cell at ix,iy
        volume(left,right)
        Volume(in decibel): k1*(k2-distance from eye), k1=1, k2=0
        :ix: cell ix
        :iy: cell iy
        :eye_xy_l: left eye/ear at x,y default: self.eye_xy_l
        :eye_xy_r: right eye/ear at x,y  default: self.eye_xy_r
        return: volume(left,right) in decibel
        """
        return self.fte.get_vol(ix=ix, iy=iy, eye_ixy_l=eye_ixy_l, eye_ixy_r=eye_ixy_r)

    def set_xy(self, xy):
        """ Set our internal win x,y
        :xy: x,y tuple as new win coordinates
        :returs: (x,y) tuple
        """
        return self.fte.set_xy(xy)

    def get_xy(self):
        """ Get current xy pair
        :returns: (x,y) tuple
        """
        return self.fte.get_xy() 

    def get_xy_canvas(self, xy=None):
        """ Get current xy pair on canvas (0-max)
        :xy: xy tuple default: current xy
        :returns: (x,y) tuple
        """
        return self.fte.get_xy_canvas(xy=xy)
    
    def set_win_fract(self, val):
        """ Set fraction indicator
        :val: new win_fract
        :returns: new win_fract
        """
        return self.fte.set_win_fract(val)

    def get_win_fract(self):
        return self.fte.get_win_fract()
    
    def set_x_min(self, val):
        """ Set min
        :val: new x_min
        :returns: new x_min
        """
        return self.fte.set_x_min(val)

    def get_x_min(self):
        return self.fte.get_x_min()

    def set_x_max(self, val):
        """ Set max
        :val: new x_max
        :returns: new x_min
        """
        return self.fte.set_x_max(val)

    def get_x_max(self):
        return self.fte.get_x_max()
        
 
    def set_y_min(self, val):
        """ Set min
        :val: new x_min
        :returns: new x_min
        """
        return self.fte.set_y_min(val)

    def get_y_min(self):
        return self.fte.get_y_min()

    def set_y_max(self, val):
        """ Set max
        :val: new y_max
        :returns: new x_min
        """
        return self.fte.set_y_max(val)

    def get_y_max(self):
        return self.fte.get_x_max()

                
    def pos_check(self, x=None, y=None, force_output=False, with_voice=False):
        """ Do position checking followed by report queue processing
        :x: x postion default: current
        :y: y position default: current
        :force_output: force output, flushing current queue
        :with_voice: say if, even if beep enable
        """
        self.fte.pos_check(x=x, y=y, force_output=force_output,
                             with_voice=with_voice)

    def is_silent(self):
        return self.fte.is_silent()


if __name__ == "__main__":
    SlTrace.setupLogging(logToScreen=True)
    def test_window_ops():
        SlTrace.clearFlags()
        app = wx.App()
        
        
        menu_str = "n:sh;d:d"
        key_str = (
                    "d"
                    ";c;g;9;9;9;9"
                    ";c;r;7;7;7;7"
                    ";c;v;2;2;2;c;r;2;2;c;o;2;2;2"
                    ";u;8;8;8;8;8;8;8;8;d"
                    ";c;o;1;1;1;1"
                    ";c;b;3;3;3;3"
                    ";c;g;6;6;c;i;6;6;6;c;v;6;6;6"
                    ";w"
                )

        key_str = (
                    "d"
                    ";c;g;9;9;9;9"
                    ";c;r;7;7;7;7"
    
                    ";w"
                )
        
        key_str = ""
        
        tkr = TkRPCUser(simulated=True)
        aw = AudioDrawWindow(tkr=tkr,
                            id_title = "Selftest-no turtle",
                            title="AudioDrawWindow Self-Test",
                            app=app,
                            menu_str=menu_str,
                            key_str=key_str,
                            silent=False)
        '''
        aw.fte.do_menu_str("n:nu;d:s")
        aw.fte.do_key_str("u")
        '''
        
        aw.MainLoop()

    def test_tk_track(figure=1):
        SlTrace.clearFlags()
        app = wx.App()
        SlTrace.lg("After clearFlags")
        SlTrace.lg("Checking talk_cmd", "talk_cmd")
        tkr = TkRPCUser(simulated=True, figure=figure)
        aw = AudioDrawWindow(tkr=tkr,
                            title="AudioDrawWindow Self-Test",
                            app=app,
                            silent=False)
        
        #SlTrace.setFlags("refresh,draw_rect,draw_oval,mouse_cell")
        aw.MainLoop()
    
    test_window_ops()
    #test_tk_track(figure=1)
    #test_tk_track(figure=2)



# wx_grid_path.py   11Jul2024  crs, convert to wx...
"""
Determine predictive following moves
"""
from select_trace import SlTrace

from braille_cell import BrailleCell

class Cell:
    @classmethod
    def from_tuple(cls, *tp):
        if len(tp) == 1:
            tp = tp[0]
            if tp is None:
                return tp
            
            if isinstance(tp, BrailleCell):
                return Cell(ix=tp.ix, iy=tp.iy)
            
            if isinstance(tp, Cell):
                return tp
            
            if isinstance(tp, tuple):
                return Cell(ix=tp[0], iy=tp[1])
            raise Exception(f"tp: {tp} unexpected type")
        
        if len(tp) == 2:
            return Cell(ix=tp[0], iy=tp[1])
        
        raise Exception(f"tp: {tp} unexpected type")
    
    def __init__(self, ix, iy):
        self.ix, self.iy = ix,iy



class GridPath:
    def __init__(self, awin, pred_len=1, det_len=4, hv_len=2):
        """ Setup path determination
        :awin: AudioDrawWin basis
        :pred_len: maximum length of predictive cells (ix,iy)
                    default: 3
        :det_len: length of goto path used to determing prediction - usually
                direction of predictive cells
                default: 4
        :hv_len: length after which line is considered horizontal/vertical
        """
        self.awin = awin
        self.pred_len = pred_len
        self.det_len = det_len
        self.hv_len = hv_len

    def get_next_position(self):
        """ get likely next move
        :returns: (ix,iy) of likely next move
        """
        pos_history = self.awin.pos_history
        if len(pos_history) < 2:
            return      # Can't tell
        
        pos_now = pos_history[-1]
        ix_now, iy_now = pos_now
        pos_prev = pos_history[-2]
        ix_prev, iy_prev = pos_prev
        ix_chg = ix_now - ix_prev
        iy_chg = iy_now - iy_prev
        ix_new = ix_now + ix_chg
        iy_new = iy_now + iy_chg
        pos_new = (ix_new,iy_new)
        cell_new = self.awin.get_cell_at_ixy(cell_ixy=pos_new)
        SlTrace.lg(f"prev:{pos_prev} now:{pos_now}"
                   f" new:{pos_new} {cell_new}"
                   , "pos_tracking")
        return (ix_new,iy_new)
        

    def get_next_positions(self, max_len=1):
        """ get likely next moves
        :max_len: maximum length of look ahead default: 1
        :returns: list of (ix,iy) of likely next move
        """
        pos_history = self.awin.pos_history
        if len(pos_history) < 2:
            return []      # Can't tell
        
        pos_now = pos_history[-1]
        ix_now, iy_now = pos_now
        pos_prev = pos_history[-2]
        ix_prev, iy_prev = pos_prev
        ix_chg = ix_now - ix_prev
        iy_chg = iy_now - iy_prev
        nexts = []
        while len(nexts) < max_len:
            ix_new = ix_now + ix_chg
            iy_new = iy_now + iy_chg
            pos_new = (ix_new,iy_new)
            if not self.awin.is_inbounds(ix=ix_new, iy=iy_new):
                break
            nexts.append(pos_new)
            ix_now,iy_now = pos_new
        return nexts
                
    def find_predictive_list(self, pred_len=None, det_len=None, 
                             hv_len=None):
        """ Find likely next path (list of cell ixiy)
        :pred_len: maximum length of predictive cells (ix,iy)
                    default: self.pred_len(3)
        :det_len: length of goto path used to determing prediction - usually
                direction of predictive cells
                default: self.det_len(4
        :hv_len: horizontal/vertical run to force consideration as
                 horizontal/vertical
        :returns: list of cell ordered to be the likely next set of cells
        """
        if pred_len is None:
            pred_len = self.pred_len
        if det_len is None:
            det_len = self.det_len
        if hv_len is None:
            hv_len = self.hv_len
        
        path_xy = self.find_path_dir_xy(det_len=det_len, 
                             goto_path=self.awin.pos_history,
                             hv_len = hv_len)
        ##print(f"path_xy: {path_xy}")
        if path_xy == (0,0):
            return []       # No prediction if can't find direction
        
        cell = Cell.from_tuple(path_xy[-1])
        if cell is None:
            return []
        
        x_start, y_start = cell.ix, cell.iy
        x_chg, y_chg = path_xy
        x_major = True if abs(x_chg) >= abs(y_chg) else False
        pred_list = []
        if x_major:
            x_chg_norm = x_chg/abs(x_chg)
            y_chg_norm = y_chg/x_chg
        else:
            y_chg_norm = y_chg/abs(y_chg)
            x_chg_norm = x_chg/y_chg
        for i in range(1, pred_len+1):
            pred_x = int(i*x_chg_norm + x_start)
            pred_y = int(i*y_chg_norm + y_start)
            pred_list.append((pred_x,pred_y)) 
             
        return pred_list
        
    def find_path_dir_xy(self, det_len=None, 
                             goto_path=None,
                             hv_len = None):
        """ Get most recent path direction (x,y)
            1. If cells are horizontal or vertical for hv_len cells
            we return the corresponding horizontal/vertical x,y
            
        :det_len: determination length (max) used to determine slope
        :goto_path: list of most recent cells, last first
        :hv_len: length to determine horizontal/vertica
        :returns: (xchg, ychg) - (0,0) == indeterminate slope
        """
        if det_len is None:
            det_len = self.det_len
        if goto_path is None:
            goto_path = self.goto_path
        if hv_len is None:
            hv_len = self.hv_len
        if len(goto_path) < 2:
            return (0,0)        # Need at least two points
        
        cell = Cell.from_tuple(goto_path[-1])
        if cell is None:
            return 0,0          # Nothing
            
        x_end,y_end = cell.ix,cell.iy
        ##print(f"end: {x_end},{y_end}")
        if len(goto_path) >= hv_len:
            for i in range(hv_len-1):
                cell = Cell.from_tuple(goto_path[-i-2])
                if cell is None:
                    break
                x = cell.ix
                if x != x_end:
                    break
            else:
                cell = Cell.from_tuple(goto_path[-hv_len])
                if cell is None:
                    return 0,0
                
                y = cell.iy
                y_chg = y_end-y
                return(0, y_chg)
            
            for i in range(hv_len-1):
                cell = Cell.from_tuple(goto_path[-i-2])
                if cell is None:
                    return 0,0
                
                y = cell.iy
                if y != x_end:
                    break
            else:
                cell = Cell.from_tuple(goto_path[-hv_len])
                if cell is None:
                    return 0,0
                
                x = cell.ix
                x_chg = x_end-y
                return(x_chg, 0)
            
            det_len = min(det_len, len(goto_path))
            cell = Cell.from_tuple(goto_path[-det_len])
            if cell is None:
                return 0,0
            
            x = cell.ix
            x_chg = x_end-x
            y = cell.iy
            y_chg = y_end-y
            return x_chg, y_chg        
        

if __name__ == "__main__":
    from audio_draw_window import AudioDrawWindow
    awin = AudioDrawWindow(title="grid_path self test")

    print(f"Self Test: {__file__}")
    cells_set = {
            (2,6), (3,6), (4,6), (5,6), (6,6),
            (2,5),               (5,5),
            (2,4),               (5,4),
            (2,3), (3,3), (4,3), (5,3),
            (2,2), (3,2), (4,2), (5,2), (6,2),
         }
    cells = dict.fromkeys(cells_set, 0)
    goto_cells = [(2,6), (2,5), (2,4), (2,3)]
    awin.pos_history = goto_cells
    gp = GridPath(awin=awin)
    #pred_cells = gp.find_predictive_list()
    #print(f"goto_path:{goto_cells},\npred_cells:{pred_cells}")

    goto_cells = [(1,3), (2,3),
                  (1,2), (2,2), (3,2),
                                 (3,1)]
    awin.pos_history = goto_cells
    pred_pos = gp.get_next_position()
    print(f"goto_path:{goto_cells},\nnext:{pred_pos}")

        


# select_report.py    13Mar2020  crs from http://effbot.org/tkinterbook/tkinter-dialog-windows.htm

"""
Report to user with a modeless message
"""
from tkinter import Tk, Label, Frame
import tkSimpleDialog

class SelectReport(tkSimpleDialog.Dialog):

    def __init__(self, master=None, title=None, message=None, grab_set=False):
        self.master = master
        self.message = message
        self.standalone = False
        if master is not None and hasattr(master, "tk"):
            master_focus = master
        elif master is not None and hasattr(master, "mw"):
            master_focus = master.mw
        else:
            master_focus = Tk()
            self.standalone = True
        self.initial_focus = master
        ###super().__init__(master_focus, title=title, grab_set=grab_set)   # TFD remove grab_set
        super().__init__(master_focus, title=title)
        
    def body(self, master):

        Label(master, text=self.message).grid(row=0)
        return self.initial_focus # initial focus

if __name__ == "__main__":

    root = Tk()


    ###SelectReport(root, message="SelectReport Testing")
    SelectReport(message="SelectReport Testing Stanalone")



# speech_maker_cmd.py 03Oct2023  crs 
"""
Speach command
TBD - independant control of speach engine properties
"""

class SpeechMakerCmd:
    """ Command to execute
    """

    def __init__(self, cmd_type=None, msg=None, msg_type=None,
                 rate=None, volume=None):
        """ Setup command
        :cmd_type: command to execute
                "MSG" - speak message
                "CLEAR" - clear pending speech
                "QUIT" - quit operation
        :msg: text to speak SpeakText
        :msg_type: type of message
                REPORT: std reporting
                CMD: command
                ECHO: echo of user input
            default: REPORT
        :rate: speech rate default: 240
        :volume: speech volume default: .9
        """
        self.cmd_type = cmd_type
        self.msg = msg
        self.msg_type = msg_type
        if rate is None:
            rate = 240
        self.rate = rate
        if volume is None:
            volume = .9
        self.volume = volume

    def __str__(self):
        ret = f"SpeechMakerCmd {self.cmd_type}"
        if self.msg is not None:
            ret += f" {self.msg}"
        if self.msg_type is not None:
            ret += f" {self.msg_type}"
        return ret
